/*
  This file is part of Simulation-Project:2D Rigid Body Collisions.
 

    Simulation-Project:2D Rigid Body Collisions is free software: you can redistribute 
    it and/or modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Simulation-Project:2D Rigid Body Collisions is distributed in 
    the hope that it will be useful, but WITHOUT ANY WARRANTY;
     without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Simulation-Project:2D Rigid Body Collisions.  
    If not, see <http://www.gnu.org/licenses/>.
*/
package ca.uoit.csci3010u.project;


class Polygon{
	
	final double radius;
	final int lines;
	final double area;
	final double mass;
	
	Point center;
	double xVelocity;
	double yVelocity;
	
	Point[] points;
	
	double rotation;
	double angularVelocity;
	
	int lastHitShape;

	/**
	 * Creates a regular polygon with given properties:
	 * 
	 * @param c 	The center point
	 * @param n 	The number of edges/vertices
	 * @param r		The radius
	 * @param xV	The initial xVelocity
	 * @param yV	The initial yVelocity
	 * @param rot	The initial rotation (Radians)
	 * @param av	The initial angular velocity
	 */
	public Polygon(final Point c,final int n,final double r,
			final double xV,final double yV, final double rot,final double av){
		lastHitShape=-1;
		center=c;
		lines=n;
		radius=r;
		xVelocity=xV;
		yVelocity=yV;
		rotation = rot;
		angularVelocity=av;
		createPoints();
		
		double sideLength = Math.sqrt(Math.pow((points[0].x - points[1].x),2) + 
				Math.pow((points[0].y - points[1].y),2));
		
		area = 1.0/4.0 * n * Math.pow(sideLength, 2) * (1/Math.tan(Math.PI/n));
		mass = area;
		
	}
	
	/** Generate the points in space that represent the verticies
	 * of this polygon.
	 */
	public void createPoints(){
		points=new Point[lines];
		final double increment= 2.0*Math.PI/lines;
		for(int i=0;i<lines;i++){

			final double tempX=Math.cos(rotation + increment*i)*radius;
			final double tempY=Math.sin(rotation + increment*i)*radius;
			points[i]=new Point(center.x+tempX,center.y+tempY);
		}
	}
	
	public double getMaxX(){
		double max_x = 0;
		for(Point p: points){
			if(p.x > max_x){
				max_x = p.x;
			}
		}
		
		return max_x;
	}
	
	public double getMaxY(){
		double max_y = 0;
		for(Point p: points){
			if(p.y > max_y){
				max_y = p.y;
			}
		}
		
		return max_y;
	}
	
	public double getMinX(){
		double min_x = 10000;
		for(Point p: points){
			if(p.x < min_x){
				min_x = p.x;
			}
		}
		
		return min_x;
	}
	public double getMinY(){
		double min_y = 10000;
		for(Point p: points){
			if(p.y < min_y){
				min_y = p.y;
			}
		}
		
		return min_y;
	}
	
	private boolean boundingCircleCollision(Polygon s){
		final double deltaX=Math.abs(this.center.x-s.center.x);
		final double deltaY=Math.abs(this.center.y-s.center.y);
		final double shortestPath=Math.sqrt(Math.pow(deltaX, 2.0)+Math.pow(deltaY, 2.0));
		if(shortestPath<s.radius+this.radius)
			return true;
		return false;
	}
	
	private boolean shapeCollision(Polygon s){
		if(s.lines>30&&this.lines>30)
			return true;
		for(int i=0;i<this.lines;i++)
			for(int j=0;j<s.lines;j++){

				if(lineIntersects(this.points[i], this.points[i==this.lines-1?0:i+1],
						s.points[j], s.points[j==s.lines-1? 0: j+1]))
					return true;
			}
		return false;
	}
	
	/** Determine if two lines intersect */
	private boolean lineIntersects(Point start1,Point end1, Point start2, Point end2) {
		//this code taken from
		//http://groups.google.com/group/android-developers/browse_thread/thread/ea7c3b2cdcf917b4

		//First find Ax+By=C values for the two lines
		double A1 = end1.y - start1.y;
		double B1 = start1.x - end1.x;
		double C1 = A1 * start1.x + B1 * start1.y;

		double A2 = end2.y - start2.y;
		double B2 = start2.x - end2.x;
		double C2 = A2 * start2.x + B2 * start2.y;

		double det = (A1 * B2) - (A2 * B1);

		if (Math.abs(det) < 0.1) {
			// Lines are either parallel, are collinear (the exact same
			// segment), or are overlapping partially, but not fully
			// To see what the case is, check if the endpoints of one line
			// correctly satisfy the equation of the other (meaning the two
			// lines have the same y-intercept).
			// If no endpoints on 2nd line can be found on 1st, they are
			// parallel.
			// If any can be found, they are either the same segment,
			// overlapping, or two segments of the same line, separated by some
			// distance.
			// Remember that we know they share a slope, so there are no other
			// possibilities

			// Check if the segments lie on the same line
			// (No need to check both points)
			if ((A1 * start2.x) + (B1 * start2.y) == C1) {
				// They are on the same line, check if they are in the same
				// space
				// We only need to check one axis - the other will follow
				if ((Math.min(start1.x, end1.x) < start2.x)
						&& (Math.max(start1.x, end1.x) > start2.x))
					return true;

				// One end point is ok, now check the other
				if ((Math.min(start1.x, end1.x) < end2.x)
						&& (Math.max(start1.x, end1.x) > end2.x))
					return true;

				// They are on the same line, but there is distance between them
				return false;
			}

			// They are simply parallel
			return false;
		} else {
			// Lines DO intersect somewhere, but do the line segments intersect?
			double x = (B2 * C1 - B1 * C2) / det;
			double y = (A1 * C2 - A2 * C1) / det;

			// Make sure that the intersection is within the bounding box of
			// both segments
			if ((x > Math.min(start1.x, end1.x) &&
					x < Math.max(start1.x, end1.x)) &&
					(y > Math.min(start1.y, end1.y) &&
							y < Math.max(start1.y, end1.y))) {
				// We are within the bounding box of the first line segment,
				// so now check second line segment
				if ((x > Math.min(start2.x, end2.x) &&
						x < Math.max(start2.x, end2.x))
						&& (y > Math.min(start2.y, end2.y) && y < Math.max(
								start2.y, end2.y))) {
					// The line segments do intersect
					return true;
				}
			}

			// The lines do intersect, but the line segments do not
			return false;
		}
	}
	
	
	public boolean checkCollision(Polygon s){
		return boundingCircleCollision(s) ? shapeCollision(s) : false;
	}
}